import json

from django.conf import settings
from django.contrib.gis.geos import GEOSGeometry, LineString, Point, GEOSException
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from ..helpers import Routing


class RoutingViewsSetMixin:
    if 'geostore.routing' in settings.INSTALLED_APPS:
        @action(detail=True, methods=['post'], permission_classes=[IsAuthenticated])
        def route(self, request, pk=None):
            layer = self.get_object()
            callback_id = self.request.data.get('callbackid', None)

            try:
                geometry = GEOSGeometry(str(request.data.get('geom')))
                if not isinstance(geometry, LineString):
                    raise ValueError

            except (GEOSException, TypeError, ValueError):
                return Response({"error": 'Provided geometry is not valid LineString'},
                                status=status.HTTP_400_BAD_REQUEST)

            points = [Point(c, srid=geometry.srid) for c in geometry.coords]
            routing = Routing(points, layer)
            route = routing.get_route()

            if not route:
                return Response(status=status.HTTP_204_NO_CONTENT)

            way = routing.get_linestring()
            response_data = {
                'request': {
                    'callbackid': callback_id,
                    'geom': json.loads(geometry.geojson),
                },
                'route': route,
                'geom': json.loads(way.geojson)
            }

            return Response(response_data)



from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import gettext as _

from geostore.models import Layer
from geostore.routing.helpers import Routing


class Command(BaseCommand):
    help = 'Update pgRouting topology'

    def add_arguments(self, parser):
        parser.add_argument('-pk',
                            '--layer-pk',
                            type=int,
                            action="store",
                            required=True,
                            help=("PK of the layer where to insert"
                                  "the features.\n"
                                  ))
        parser.add_argument('-t',
                            '--tolerance',
                            type=float,
                            action="store",
                            required=False,
                            help=_("Tolerance for snapping topologies (default to 0.00001)."))

    @transaction.atomic()
    def handle(self, *args, **options):
        layer_pk = options.get('layer_pk', None)
        tolerance = options.get('tolerance', 0.00001)

        try:
            layer = Layer.objects.get(pk=layer_pk)
            if Routing.create_topology(layer, tolerance=tolerance):
                if options['verbosity'] >= 1:
                    self.stdout.write('Topology successfully updated')

            else:
                raise CommandError('An error occuring during topology update')

        except Layer.DoesNotExist:
            raise CommandError(f"Layer with pk {layer_pk} doesn't exist")



import json

from django.contrib.gis.geos import GEOSGeometry, MultiLineString, LineString
from django.core.cache import cache
from django.db import connection
from django.db.models import F, Value

from ..tiles.funcs import ST_Distance, ST_LineLocatePoint, ST_LineSubstring


def cached_segment(func, expiration=3600 * 24):
    def wrapper(self, from_point, to_point, *args, **kwargs):
        cache_key = (f'route_{self.layer.pk}'
                     f'_segment_{from_point.pk}_{from_point.fraction}'
                     f'_{to_point.pk}_{to_point.fraction}')

        def build_segment():
            return func(self, from_point, to_point, *args, **kwargs)

        return cache.get_or_set(cache_key, build_segment, expiration)

    return wrapper


class Routing(object):

    def __init__(self, points, layer):
        if not layer.is_linestring or layer.is_multi:
            raise ValueError('Layer is not routable')

        self.layer = layer
        self.points = self._get_points_in_lines(points)
        self.routes = self._points_route()

    def get_route(self):
        """ Return the geometry of the route from the given points """

        if self.routes:
            return self._serialize_routes(self.routes)

    def get_linestring(self):
        if self.routes:
            way = MultiLineString(*[GEOSGeometry(route['geometry'])
                                    for route in self.routes if isinstance(GEOSGeometry(route['geometry']),
                                                                           LineString)])
            return way.merged

    @classmethod
    def create_topology(cls, layer, tolerance=0.00001, clean=False):
        cursor = connection.cursor()
        raw_query = """
                    SELECT
                        pgr_createTopology(
                            %s,
                            %s,
                            'geom',
                            'id',
                            'source',
                            'target',
                            rows_where := %s,
                            clean := %s)
                    """

        cursor.execute(raw_query,
                       [layer.features.model._meta.db_table, tolerance, f'layer_id={layer.pk} ', clean])
        return ('OK',) == cursor.fetchone()

    def _serialize_routes(self, routes):
        return {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry":
                        json.loads(GEOSGeometry(route['geometry']).geojson),
                    "properties": route['properties'],
                }
                for route in routes
            ]
        }

    def _get_points_in_lines(self, points):
        '''Returns position of the point in the closed geometry'''
        snapped_points = []

        for point in points:
            closest_feature = self._get_closest_geometry(point)

            snapped_points.append(
                self._snap_point_on_feature(point, closest_feature))

        return snapped_points

    def _get_closest_geometry(self, point):
        return self.layer.features.all().annotate(
            distance=ST_Distance(F('geom'), Value(str(point)))
        ).order_by('distance').first()

    def _snap_point_on_feature(self, point, feature):
        return self.layer.features.annotate(
            fraction=ST_LineLocatePoint(F('geom'), Value(str(point))),
        ).get(pk=feature.pk)

    def _points_route(self):
        route = []
        for index, point in enumerate(self.points):
            try:
                next_point = self.points[index + 1]
            except IndexError:
                break

            segment = self._get_segment(point, next_point)

            if segment:
                route.extend(segment)

        return route

    @cached_segment
    def _get_segment(self, from_point, to_point):
        if from_point.pk == to_point.pk:
            # If both points are on same edge we do not need pgRouting
            # just split the edge from point to point.
            segment = [self._get_line_substring(from_point,
                                                [from_point.fraction,
                                                 to_point.fraction]), ]
        else:
            # Ask pgRouting the route from point to the next point
            segment = self._get_raw_route(from_point, to_point)

        return segment

    def _get_line_substring(self, feature, fractions):
        feature = self.layer.features.annotate(
            splitted_geom=ST_LineSubstring(F('geom'),
                                           float(min(fractions)),
                                           float(max(fractions)))
        ).get(pk=feature.pk)

        return {
            'geometry': feature.splitted_geom,
            'properties': feature.properties,
        }

    def _get_raw_route(self, start_point, end_point):
        """Return raw route between two points from pgrouting's
        pgr_withPoints function that need to be transformed to
        real geometry.
        """

        q = """
        WITH points AS (
            -- This is a virtual table of points used for routing and
            -- their position on the closest edge.
            SELECT
                points.pid,
                points.edge_id,
                ST_LineSubstring(geostore_feature.geom,
                                 points.fraction_start,
                                 points.fraction_end) AS geom
            FROM
                (VALUES
                    (1, %s, 0, %s::float),
                    (1, %s, %s::float, 1),
                    (2, %s, 0, %s::float),
                    (2, %s, %s::float, 1)
                ) AS points(pid, edge_id, fraction_start, fraction_end)
                JOIN geostore_feature ON
                    geostore_feature.id = points.edge_id
        ),
        pgr AS (
            -- Here we do the routing from point 1 to point 2 using
            -- pgr_withPoints that uses the dijkstra algorythm. next_node
            -- and next_geom are used later to reconstruct the final
            -- geometry of the shortest path.
            SELECT
                pgr.path_seq,
                pgr.node,
                pgr.edge,
                geostore_feature.geom AS edge_geom,
                geostore_feature.properties,
                (LEAD(pgr.node) OVER (ORDER BY path_seq))
                    AS next_node,
                (LAG(geostore_feature.geom) OVER (ORDER BY path_seq))
                    AS prev_geom,
                (LEAD(geostore_feature.geom) OVER (ORDER BY path_seq))
                    AS next_geom
            FROM
                pgr_withPoints(
                    'SELECT id, source, target,
                        ST_Length(geom) AS cost,
                        ST_Length(geom) AS reverse_cost
                        FROM geostore_feature
                        WHERE layer_id = %s
                            AND source IS NOT NULL
                        ORDER BY id',
                    'SELECT *
                        FROM (VALUES (1, %s, %s::float), (2, %s, %s::float))
                        AS points (pid, edge_id, fraction)',
                    -1, -2, details := true
                ) AS pgr
            LEFT OUTER JOIN geostore_feature ON pgr.edge = geostore_feature.id
        ),
        route AS (
            /* Finally we reconstruct the geometry by collecting each edge.
                At point 1 and 2, we get the split edge.
            */
            SELECT
                CASE
                WHEN node = -1 THEN  -- Start Point
                    (SELECT points.geom
                        FROM points
                        WHERE points.pid = -pgr.node
                        -- Non topological graph,
                        -- get the closest to the next edge
                        ORDER BY ST_Distance(points.geom, pgr.next_geom) ASC
                        LIMIT 1)
                WHEN next_node = -2 THEN  -- Going to End Point
                    (SELECT points.geom
                        FROM points
                        WHERE points.pid = -pgr.next_node
                        -- Non topologic graph,
                        -- get the closest to the previous egde
                        ORDER BY ST_Distance(points.geom, pgr.prev_geom) ASC
                        LIMIT 1)
                ELSE
                    edge_geom  -- Let's return the full edge geometry
                END AS final_geometry,
                properties
            FROM pgr
        )
        SELECT
            final_geometry AS geometry,
            properties
        FROM route
        WHERE final_geometry IS NOT NULL;
        """
        self._fix_point_fraction(start_point)
        self._fix_point_fraction(end_point)

        with connection.cursor() as cursor:
            cursor.execute(q, [
                start_point.pk, float(start_point.fraction),
                start_point.pk, float(start_point.fraction),
                end_point.pk, float(end_point.fraction),
                end_point.pk, float(end_point.fraction),
                self.layer.pk,
                start_point.pk, float(start_point.fraction),
                end_point.pk, float(end_point.fraction),
            ])

            return [
                {
                    'geometry': geometry,
                    'properties': properties
                } for geometry, properties in cursor.fetchall()
            ]

    def _fix_point_fraction(self, point):
        """ This function is used to fix problem with pgrouting when point
        position on the edge is 0.0 or 1.0, that create a routing topology
        problem. See https://github.com/pgRouting/pgrouting/issues/760
        So we create a fake fraction near the vertices of the edge.
        """
        if float(point.fraction) == 1.0:
            point.fraction = 0.99999
        elif float(point.fraction) == 0.0:
            point.fraction = 0.00001



from django.conf import settings
from django.urls import reverse
from rest_framework import serializers


class RoutingLayerSerializer:
    if 'geostore.routing' in settings.INSTALLED_APPS:
        routing_url = serializers.SerializerMethodField()

        def get_routing_url(self, obj):
            return reverse('layer-route', args=[obj.pk, ])



class SIGTools(object):
    """
    This class collects all the sig functions  to use in TileVect, GeomConfig
    """

    @classmethod
    def get_extent_of_layer(cls, layer):
        """
        Outputs length of the smaller size of the features' bbox
        Returns 0 if the bbox is a Point
        """

        min_extent_features = float(0)

        # output might be a Point if single Point feature
        query = layer.get_extent()
        extent = query['extent']

        if extent:
            x1 = query['extent'][0]
            x2 = query['extent'][2]
            y1 = query['extent'][1]
            y2 = query['extent'][3]

            if x1 != x2 and y1 != y2:
                # not a point
                min_extent_features = min(float(abs(x2 - x1)), float(abs(y2 - y1)))

        return min_extent_features



from django.db import transaction

from geostore import settings as app_settings
from geostore.tasks import feature_update_relations_destinations, layer_relations_set_destinations


def execute_async_func(async_func, args=()):
    """ Celery worker can be out of transaction, and raise DoesNotExist """
    async_func.delay(*args) if not transaction.get_connection().in_atomic_block else transaction.on_commit(
        lambda: async_func.delay(*args))


def save_feature(sender, instance, **kwargs):
    if app_settings.GEOSTORE_RELATION_CELERY_ASYNC:
        execute_async_func(feature_update_relations_destinations, (instance.pk,))


def save_layer_relation(sender, instance, **kwargs):
    if app_settings.GEOSTORE_RELATION_CELERY_ASYNC:
        execute_async_func(layer_relations_set_destinations, (instance.pk,))



# Generated by Django 2.2.6 on 2019-11-15 12:56

import django.contrib.gis.db.models.fields
try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.db import migrations, models
import django.db.models.deletion
import geostore
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0032_auto_20191016_0844'),
    ]

    operations = [
        migrations.CreateModel(
            name='LayerExtraGeom',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('settings', JSONField(blank=True, default=dict)),
                ('geom_type', models.IntegerField(choices=[(0, geostore.GeometryTypes(0)), (1, geostore.GeometryTypes(1)), (3, geostore.GeometryTypes(3)), (4, geostore.GeometryTypes(4)), (5, geostore.GeometryTypes(5)), (6, geostore.GeometryTypes(6)), (7, geostore.GeometryTypes(7))], null=True)),
                ('slug', models.SlugField(editable=False)),
                ('title', models.CharField(max_length=250)),
                ('layer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extra_geometries', to='geostore.Layer')),
            ],
            options={
                'unique_together': {('layer', 'title'), ('layer', 'slug')},
            },
        ),
        migrations.CreateModel(
            name='FeatureExtraGeom',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('geom', django.contrib.gis.db.models.fields.GeometryField(srid=4326)),
                ('properties', JSONField(blank=True, default=dict)),
                ('identifier', models.UUIDField(blank=True, default=uuid.uuid4, editable=False, null=True)),
                ('feature', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extra_geometries', to='geostore.Feature')),
                ('layer_extra_geom', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='features', to='geostore.LayerExtraGeom')),
            ],
            options={
                'unique_together': {('feature', 'layer_extra_geom')},
            },
        ),
    ]



from django.urls import path, include
from rest_framework import routers

from . import views

router = routers.DefaultRouter()

router.register(r'layer', views.LayerViewSet)
router.register(r'group', views.LayerGroupViewsSet, basename='group'),
router.register(r'layer/(?P<layer>[\d\w\-_]+)/feature', views.FeatureViewSet,
                basename='feature')

urlpatterns = [
    path('', include(router.urls)),
]



# Generated by Django 3.0.6 on 2020-06-24 14:40

import django.contrib.gis.db.models.fields
import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0036_auto_20200116_0926'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='feature',
            name='geostore_fe_identif_f6c552_idx',
        ),
        migrations.AlterField(
            model_name='featureextrageom',
            name='geom',
            field=django.contrib.gis.db.models.fields.GeometryField(spatial_index=False, srid=4326),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['layer', 'identifier'], name='geostore_fe_layer_i_141c81_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=django.contrib.postgres.indexes.GinIndex(fields=['properties'], name='properties_gin_index'),
        ),
        migrations.AddIndex(
            model_name='featureextrageom',
            index=models.Index(fields=['layer_extra_geom', 'identifier'], name='geostore_fe_layer_e_aa3cea_idx'),
        ),
        migrations.AddIndex(
            model_name='featureextrageom',
            index=django.contrib.postgres.indexes.GistIndex(fields=['layer_extra_geom', 'geom'], name='feg_geom_gist_index'),
        ),
        migrations.AddIndex(
            model_name='featureextrageom',
            index=django.contrib.postgres.indexes.GinIndex(fields=['properties'], name='feg_properties_gin_index'),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-07 14:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0029_auto_20190926_0803'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='layer',
            options={'ordering': ['id'], 'permissions': (('can_manage_layers', 'Has all permissions on layers'), ('can_export_layers', 'Is able to export layers'), ('can_import_layers', 'Is able to import layers'))},
        ),
    ]



import gpxpy.gpx
import simplekml
from django.conf import settings
from django.contrib.gis.geos import GeometryCollection, Point, LineString, Polygon
from rest_framework.renderers import JSONRenderer, BaseRenderer
from rest_framework.utils.serializer_helpers import ReturnDict

from geostore.models import Feature


class GeoJSONRenderer(JSONRenderer):
    """
    GeoJSON renderer is just used to match with geojson format.
    Serializer is adapted along this format with djangorestframework-gis
    """
    format = 'geojson'


class KMLRenderer(BaseRenderer):
    format = 'kml'
    media_type = 'application/vnd.google-earth.kml+xml'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kml = simplekml.Kml()

    def get_element_infos(self, element):
        description = ""
        for key, value in element.get('properties').items():
            description += f"{key}: {value}\n"
        return (
            element.get('geom').get('type').lower(),
            element.get('identifier'),
            element.get('geom').get('coordinates'),
            description
        )

    def parse_element(self, element):
        mapping = {
            'point': 'newpoint',
            'linestring': 'newlinestring',
            'polygon': 'newpolygon'
        }
        geom_type, identifier, coordinates, description = self.get_element_infos(element)

        if geom_type in mapping.keys():
            if geom_type == "polygon":
                polygon = self.kml.newpolygon(name=identifier, description=description)
                polygon.outerboundaryis = (coordinates[0], )
                if len(coordinates) > 1:
                    # manage holes
                    polygon.innerboundaryis = (coordinates[1], )
            else:
                getattr(self.kml, mapping[geom_type])(
                    name=identifier,
                    description=description,
                    coords=(coordinates, ) if geom_type == "point" else coordinates
                )
        else:
            # geom is multi
            multi = self.kml.newmultigeometry(name=identifier, description=description)
            final_geom_type = geom_type.replace('multi', '')  # get geom_type without multi
            if final_geom_type == "geometrycollection":
                geometries = element.get('geom').get('geometries')
                for geometry in geometries:
                    geom_type = geometry.get('type').lower()
                    coordinates = geometry.get('coordinates')
                    getattr(multi, mapping[geom_type])(
                        coords=(coordinates, ) if geom_type == "point" else coordinates
                    )
            else:
                for simple_object_coords in coordinates:
                    if final_geom_type == "polygon":
                        polygon = multi.newpolygon(name=identifier, description=description)
                        polygon.outerboundaryis = (simple_object_coords[0], )
                        if len(simple_object_coords) > 1:
                            # manage holes
                            polygon.innerboundaryis = (simple_object_coords[1], )
                    else:
                        getattr(multi, mapping[final_geom_type])(
                            coords=(simple_object_coords, ) if final_geom_type == "point" else simple_object_coords
                        )

        return self.kml.kml()

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if isinstance(data, ReturnDict):
            # simple object serialization
            self.parse_element(data)

        else:
            # object list serialization
            for element in data:
                self.parse_element(element)

        return self.kml.kml()


class GPXRenderer(BaseRenderer):
    format = 'gpx'
    media_type = 'application/gpx+xml'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gpx = gpxpy.gpx.GPX()

    def render(self, data, accepted_media_type=None, renderer_context=None):
        feat = Feature.objects.get(identifier=data.get('identifier'))
        return self.geom_to_gpx(feat.geom, feat.identifier, "")

    def _point_to_gpx(self, point, klass=gpxpy.gpx.GPXWaypoint):
        if isinstance(point, (tuple, list)):
            point = Point(*point, srid=settings.INTERNAL_GEOMETRY_SRID)
        newpoint = point.transform(4326, clone=True)  # transformation: gps uses 4326
        # transform looses the Z parameter
        return klass(latitude=newpoint.y, longitude=newpoint.x, elevation=point.z)

    def geom_to_gpx(self, geom, name, description):
        """Convert a geometry to a gpx entity.
        Point -> add as a Way Point
        LineString -> add all Points in a Route
        Polygon -> add all Points of the external linering in a Route
        Collection (of LineString or Point) -> add as a route, concatening all points
        """
        if isinstance(geom, GeometryCollection):
            for i, g in enumerate(geom):
                self.geom_to_gpx(g, "%s (%s)" % (name, i), description)
        elif isinstance(geom, Point):
            wp = self._point_to_gpx(geom)
            wp.name = name
            wp.description = description
            self.gpx.waypoints.append(wp)
        elif isinstance(geom, LineString):
            gpx_track = gpxpy.gpx.GPXTrack(name=name, description=description)
            gpx_segment = gpxpy.gpx.GPXTrackSegment()
            gpx_segment.points = [self._point_to_gpx(point, klass=gpxpy.gpx.GPXTrackPoint) for point in geom]
            gpx_track.segments.append(gpx_segment)
            self.gpx.tracks.append(gpx_track)
        elif isinstance(geom, Polygon):
            self.geom_to_gpx(geom[0], name, description)

        return self.gpx.to_xml()



from django.db import models


class PgRoutingMixin(models.Model):
    source = models.IntegerField(null=True,
                                 blank=True,
                                 help_text='Internal field used by pgRouting',
                                 editable=False)
    target = models.IntegerField(null=True,
                                 blank=True,
                                 help_text='Internal field used by pgRouting',
                                 editable=False)

    class Meta:
        abstract = True



# Generated by Django 2.2.6 on 2019-10-16 08:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0031_layer_authorized_groups'),
    ]

    operations = [
        migrations.AlterField(
            model_name='layer',
            name='authorized_groups',
            field=models.ManyToManyField(blank=True, related_name='authorized_layers', to='auth.Group'),
        ),
    ]



import json
from copy import deepcopy

from django.contrib.gis.gdal.error import GDALException
from django.contrib.gis.geos import GEOSException, GEOSGeometry
from django.core.serializers import serialize
from django.db import transaction
from django.db.models import Q
from django.http import HttpResponse, HttpResponseBadRequest
from django.utils.datastructures import MultiValueDictKeyError
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404
from rest_framework.permissions import IsAuthenticated
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
from rest_framework.response import Response

from geostore.renderers import KMLRenderer, GPXRenderer
from .mixins import MultipleFieldLookupMixin
from ..filters import JSONFieldFilterBackend, JSONFieldOrderingFilter, JSONSearchField
from ..models import Layer, LayerGroup
from ..permissions import FeaturePermission, LayerPermission, LayerImportExportPermission
from ..renderers import GeoJSONRenderer
from ..routing.views.mixins import RoutingViewsSetMixin
from ..serializers import (FeatureExtraGeomSerializer, FeatureSerializer,
                           LayerSerializer)
from ..serializers.geojson import FinalGeoJSONSerializer
from ..tiles.mixins import MVTViewMixin, MultipleMVTViewMixin


class LayerGroupViewsSet(MultipleMVTViewMixin, viewsets.ReadOnlyModelViewSet):
    queryset = LayerGroup.objects.all()
    lookup_field = 'slug'


class LayerViewSet(MultipleFieldLookupMixin, MVTViewMixin, RoutingViewsSetMixin, viewsets.ModelViewSet):
    permission_classes = (LayerPermission, )
    queryset = Layer.objects.all()
    serializer_class = LayerSerializer
    lookup_fields = ('pk', 'name')

    @action(methods=['get', 'post'],
            url_name='shapefile', detail=True, permission_classes=[IsAuthenticated,
                                                                   LayerImportExportPermission])
    def shapefile(self, request, *args, **kwargs):
        layer = self.get_object()

        if request.method == 'POST':
            try:
                shape_file = request.data['shapefile']
                with transaction.atomic():
                    layer.features.all().delete()
                    layer.from_shapefile(shape_file)
                    response = Response(status=status.HTTP_200_OK)

            except (ValueError, MultiValueDictKeyError):
                response = Response(status=status.HTTP_400_BAD_REQUEST)

        else:
            shape_file = layer.to_shapefile()

            if shape_file:
                response = HttpResponse(content_type='application/zip')
                response['Content-Disposition'] = (
                    'attachment; '
                    f'filename="{layer.name}.zip"')

                response.write(shape_file.getvalue())
            else:
                response = Response(status=status.HTTP_204_NO_CONTENT)

        return response

    @action(detail=True, methods=['post'], permission_classes=[IsAuthenticated])
    def intersects(self, request, *args, **kwargs):
        layer = self.get_object()
        callbackid = self.request.data.get('callbackid', None)

        try:
            geometry = GEOSGeometry(request.data.get('geom', None))
        except (GEOSException, GDALException, TypeError, ValueError):
            return HttpResponseBadRequest(
                content='Provided geometry is not valid')

        response = {
            'request': {
                'callbackid': callbackid,
                'geom': geometry.json,
            },
            'results': json.loads(serialize('geojson',
                                            layer.features.intersects(geometry),
                                            fields=('properties',),
                                            geometry_field='geom',
                                            properties_field='properties')),
        }

        return Response(response)

    def partial_update(self, request, *args, **kwargs):
        layer = self.get_object()

        if 'features' in request.data:
            try:
                features = layer.update_geometries(request.data['features'])
                return Response(
                    json.loads(serialize('geojson',
                                         features,
                                         fields=('properties',),
                                         geometry_field='geom',
                                         properties_field='properties'))
                )
            except (ValueError, KeyError):
                return HttpResponseBadRequest('An error occured parsing '
                                              'GeoJSON, verify your data')
        else:
            return HttpResponseBadRequest('Features are missing in GeoJSON')


class FeatureViewSet(viewsets.ModelViewSet):
    permission_classes = (FeaturePermission, )
    serializer_class = FeatureSerializer
    serializer_class_extra_geom = FeatureExtraGeomSerializer
    renderer_classes = (JSONRenderer, GeoJSONRenderer, BrowsableAPIRenderer, KMLRenderer, GPXRenderer)
    filter_backends = (JSONFieldFilterBackend, JSONFieldOrderingFilter, JSONSearchField)
    filter_fields = ('properties', )
    ordering_fields = ('id', 'identifier', 'created_at', 'updated_at')
    lookup_field = 'identifier'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.layer = None

    def transform_serializer_geojson(self, serializer_class):
        if self.kwargs.get('format', 'json') == 'geojson':
            # auto override in geojson case
            class FinalClass(FinalGeoJSONSerializer, serializer_class):
                class Meta(FinalGeoJSONSerializer.Meta, serializer_class.Meta):
                    pass
            return FinalClass
        return serializer_class

    def get_serializer_class(self):
        original_class = super().get_serializer_class()
        return self.transform_serializer_geojson(original_class)

    def get_layer(self):
        if not self.layer:
            filters = Q(name=self.kwargs.get('layer'))
            if self.kwargs.get('layer').isdigit():
                filters |= Q(pk=self.kwargs.get('layer'))

            self.layer = get_object_or_404(Layer, filters)
        return self.layer

    def get_serializer_context(self):
        """
        Layer access in serializer (pk to insure schema generation)
        """
        context = super().get_serializer_context()
        layer = self.get_layer()
        context.update({'layer_pk': layer.pk})
        return context

    def get_queryset(self):
        layer = self.get_layer()
        qs = layer.features.all()
        qs = qs.prefetch_related('layer__relations_as_origin')
        return qs

    def perform_create(self, serializer):
        layer = self.get_layer()
        serializer.save(layer_id=layer.pk)

    def update(self, request, *args, **kwargs):
        """ override to keep unfilled properties in partial update case """
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        data = request.data
        if partial:
            # if partial, we update properties
            properties = deepcopy(instance.properties)
            properties.update(data.get('properties', {}))
            data['properties'] = properties
        serializer = self.get_serializer(instance, data=data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response(serializer.data)

    @action(detail=True, methods=['get', 'put', 'patch', 'delete'],
            url_path=r'extra_geometry/(?P<id_extra_feature>\d+)', url_name='detail-extra-geometry')
    def extra_geometry(self, request, id_extra_feature, *args, **kwargs):
        feature = self.get_object()
        extra_geometry = get_object_or_404(feature.extra_geometries.all(), pk=id_extra_feature)
        extra_layer = extra_geometry.layer_extra_geom
        if request.method == 'GET':
            return Response(self.serializer_class_extra_geom(extra_geometry).data)
        if not extra_layer.editable:
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
        if request.method == 'DELETE':
            extra_geometry.delete()
            return Response(status=status.HTTP_204_NO_CONTENT)
        elif request.method in ('PUT', 'PATCH'):
            serializer = self.serializer_class_extra_geom(data=request.data, instance=extra_geometry)
            if serializer.is_valid():
                serializer.save()
            else:
                return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
            return Response(serializer.data)

    @action(detail=True, methods=['post'], url_path=r'extra_layer/(?P<id_extra_layer>\d+)',
            url_name='create-extra-geometry')
    def extra_layer_geometry(self, request, id_extra_layer, *args, **kwargs):
        feature = self.get_object()
        layer = self.get_layer()
        extra_layer = get_object_or_404(layer.extra_geometries.all(), pk=id_extra_layer)
        if not extra_layer.editable:
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)
        serializer = self.serializer_class_extra_geom(data=request.data)
        if serializer.is_valid():
            serializer.save(feature=feature, layer_extra_geom=extra_layer)
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    @action(detail=True, url_path=r'relation/(?P<id_relation>[\d-]+)/features')
    def relation(self, request, *args, **kwargs):
        feature = self.get_object()
        layer_relation = get_object_or_404(feature.layer.relations_as_origin.all(),
                                           pk=kwargs.get('id_relation'))
        qs = feature.get_stored_relation_qs(layer_relation)
        # keep original viewset filtering
        qs = self.filter_queryset(qs)
        # keep original viewset pagination
        page = self.paginate_queryset(qs)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(qs, many=True)
        return Response(serializer.data)



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0001_initial'),
    ]

    operations = [
        migrations.RunSQL("CREATE EXTENSION IF NOT EXISTS pgrouting")
    ]



# Generated by Django 3.0.7 on 2020-06-25 15:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0040_merge_20200625_1508'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='feature',
            name='geostore_fe_propert_46222e_gin',
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['source', 'target', 'layer'], name='geostore_fe_source_155396_idx'),
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-31 17:58

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0038_auto_20200128_1442'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='feature',
            index=django.contrib.postgres.indexes.GinIndex(fields=['properties'], name='geostore_fe_propert_46222e_gin'),
        ),
    ]



from django.contrib.gis.db.models import (FloatField, GeometryField,
                                          IntegerField)
from django.db.models import Func


class RawGeometryField(GeometryField):
    def select_format(self, compiler, sql, params):
        """
        Override compiler format to not cast as bytea
        """
        return sql, params


class ST_AsMvtGeom(Func):
    function = 'ST_AsMvtGeom'
    output_field = RawGeometryField()


class ST_Transform(Func):
    function = 'ST_Transform'
    output_field = RawGeometryField()


class ST_MakeEnvelope(Func):
    function = 'ST_MakeEnvelope'
    output_field = RawGeometryField()


class ST_Distance(Func):
    function = 'ST_Distance'
    output_field = FloatField()


class ST_LineLocatePoint(Func):
    function = 'ST_LineLocatePoint'
    output_field = FloatField()


class ST_LineInterpolatePoint(Func):
    function = 'ST_LineInterpolatePoint'
    output_field = RawGeometryField()


class ST_Split(Func):
    function = 'ST_Split'
    output_field = RawGeometryField()


class ST_LineSubstring(Func):
    function = 'ST_LineSubstring'
    output_field = GeometryField()


class ST_SRID(Func):
    function = 'ST_SRID'
    output_field = IntegerField()


class ST_HausdorffDistance(Func):
    function = 'ST_HausdorffDistance'
    output_field = FloatField()


class ST_SnapToGrid(Func):
    function = 'ST_SnapToGrid'
    output_field = RawGeometryField()


class ST_Buffer(Func):
    function = 'ST_Buffer'
    output_field = RawGeometryField()


class ST_SetEffectiveArea(Func):
    function = 'ST_SetEffectiveArea'
    output_field = RawGeometryField()


class ST_Length(Func):
    function = 'ST_Length'
    output_field = FloatField()


class ST_Area(Func):
    function = 'ST_Area'
    output_field = FloatField()


class ST_MakeValid(Func):
    function = 'ST_MakeValid'
    output_field = RawGeometryField()


class ST_CollectionExtract(Func):
    function = 'ST_CollectionExtract'
    output_field = RawGeometryField()


class ST_SimplifyPreserveTopology(Func):
    function = 'ST_SimplifyPreserveTopology'
    output_field = RawGeometryField()
